from JciHitachi.api import AWSThing
from JciHitachi.model import STATUS_DICT, JciHitachiAWSStatus, JciHitachiAWSStatusSupport


class TestStatusDict:
    device_types = AWSThing.supported_device_type.values()
    
    def test_sanity(self):    
        attribute_keys = {"controllable", "is_numeric", "legacy_name"}

        for device_type in self.device_types:
            for status_name, status_attributes in STATUS_DICT[device_type].items():
                # Test if attribute keys are in status attributes.
                assert len(attribute_keys - set(status_attributes)) == 0, \
                    f"Lack of attribute keys: {status_name} {set(status_attributes)}"
                # Test is_numeric and id2str mutual exclusiveness
                assert status_attributes["is_numeric"] ^ ("id2str" in status_attributes), \
                    f"is_numeric and id2str are mutually exclusive: {status_name}"
        
class TestModel:
    def test_ac_correctness(self):
        raw_ac_status = {
            'DeviceType': 1,
            'Switch': 0,
            'Mode': 0,
            'FanSpeed': 3,
            'TemperatureSetting': 26,
            'IndoorTemperature': 23,
            'SleepModeRemainingTime': 0,
            'MildewProof': 0,
            'QuickMode': 0,
            'PowerSaving': 0,
            'ControlTone': 0,
            'PowerConsumption': 5,
            'TaiseiaError': 0,
            'FilterElapsedHour': 800,
            'CleanSwitch': 0,
            'CleanNotification': 0,
            'CleanStatus': 0,
            'Error': 0,
            'RequestTimestamp': 1648618952,
            'Timestamp': 1648618953
        }
        processed_ac_status = {
            'CleanNotification': 0,
            'CleanStatus': 0,
            'CleanSwitch': 'off',
            'ControlTone': 'enabled',
            'DeviceType': 'AC',
            'Error': 0,
            'FanSpeed': 'moderate',
            'FilterElapsedHour': 800,
            'IndoorTemperature': 23,
            'MildewProof': 'disabled',
            'Mode': 'cool',
            'PowerConsumption': 0.5,
            'PowerSaving': 'disabled',
            'QuickMode': 'disabled',
            'SleepModeRemainingTime': 0,
            'Switch': 'off',
            'TaiseiaError': 0,
            'TemperatureSetting': 26,
        }

        status = JciHitachiAWSStatus(raw_ac_status)
        assert status.status == processed_ac_status
        assert len(processed_ac_status) == len(raw_ac_status) - 2  # no `RequestTimestamp` and `Timestamp`

        raw_ac_support = {
            'FirmwareId':3,
            'DeviceType':1,
            'Model':'RAD-90NF',
            'FindMe':10,
            'Switch':3,
            'Mode':31,
            'FanSpeed':31,
            'TemperatureSetting':4128,
            'IndoorTemperature':40,
            'SleepModeRemainingTime':1440,
            'MildewProof':3,
            'QuickMode':3,
            'PowerSaving':3,
            'ControlTone':3,
            'PowerConsumption':65535,
            'TaiseiaError':85,
            'FilterElapsedHour':800,
            'CleanSwitch':3,
            'CleanNotification':3,
            'CleanStatus':7,
            'Error':0,
            'FirmwareVersion':'6.0.035',
            'FirmwareCode':35,
            'SystemTimestamp':35000,
            'RequestTimestamp':1648618952,
            'Timestamp':1648618952,
        }
        processed_ac_support = {
            'FirmwareId': 3,
            'DeviceType': 'AC',
            'Model': 'RAD-90NF',
            'FindMe': 10,
            'Switch': 3,
            'Mode': 31,
            'FanSpeed': 31,
            'TemperatureSetting': 4128,
            'IndoorTemperature': 40,
            'SleepModeRemainingTime': 1440,
            'MildewProof': 3,
            'QuickMode': 3,
            'PowerSaving': 3,
            'ControlTone': 3,
            'PowerConsumption': 65535,
            'TaiseiaError': 85,
            'FilterElapsedHour': 800,
            'CleanSwitch': 3,
            'CleanNotification': 3,
            'CleanStatus': 7,
            'Error': 0,
            'FirmwareVersion':'6.0.035',
            'FirmwareCode': 35,
            'SystemTimestamp': 35000,
            'RequestTimestamp': 1648618952,
            'Timestamp': 1648618952,
            'Brand': 'HITACHI',
            'max_temp': 32,
            'min_temp': 16
        }

        support = JciHitachiAWSStatusSupport(raw_ac_support)
        assert support.status == processed_ac_support
        assert len(processed_ac_support) == len(raw_ac_support) + 3  # `Brand` `max_temp` `min_temp`
    

    def test_he_correctness(self):
        raw_he_status = {
            'DeviceType': 3,
            'Switch': 0,
            'FanSpeed': 4,
            'IndoorTemperature': 31,
            'TaiseiaError': 0,
            'CleanFilterNotification': 0,
            'BreathMode': 0,
            'FrontFilterNotification': 0,
            'Pm25FilterNotification': 0,
            'Error': 0,
            'RequestTimestamp': 1661593958,
            'ReceiveTimestamp': 1661593958,
            'Timestamp': 1661593958
        }
        processed_he_status = {
            'DeviceType': 'HE',
            'Switch': 'off',
            'FanSpeed': 'high',
            'IndoorTemperature': 31,
            'TaiseiaError': 0,
            'CleanFilterNotification': 'disabled',
            'BreathMode': 'auto',
            'FrontFilterNotification':'disabled',
            'Pm25FilterNotification': 'disabled',
            'Error': 0
        }

        status = JciHitachiAWSStatus(raw_he_status)
        assert status.status == processed_he_status
        assert len(processed_he_status) == len(raw_he_status) - 3  # no `RequestTimestamp`, `ReceiveTimestamp``, and `Timestamp`

        raw_he_support = {
            'FirmwareId': 1,
            'DeviceType': 3,
            'Model': 'KPI-H',
            'FindMe': 10,
            'Switch': 3,
            'FanSpeed': 20,
            'IndoorTemperature': 33151,
            'TaiseiaError': 244,
            'CleanFilterNotification': 1,
            'BreathMode': 7,
            'FrontFilterNotification': 1,
            'Pm25FilterNotification': 1,
            'Error': 0,
            'FirmwareVersion': '6.0.036',
            'FirmwareCode': 52,
            'SystemTimestamp': 81274,
            'RequestTimestamp': 1661596297,
            'Timestamp': 1661596297
        }
        processed_he_support = {
            'FirmwareId': 1,
            'DeviceType': 'HE',
            'Model': 'KPI-H',
            'FindMe': 10,
            'Switch': 3,
            'FanSpeed': 20,
            'IndoorTemperature': 33151,
            'TaiseiaError': 244,
            'CleanFilterNotification': 1,
            'BreathMode': 7,
            'FrontFilterNotification': 1,
            'Pm25FilterNotification': 1,
            'Error': 0,
            'FirmwareVersion': '6.0.036',
            'FirmwareCode': 52,
            'SystemTimestamp': 81274,
            'RequestTimestamp': 1661596297,
            'Timestamp': 1661596297,
            'Brand': 'HITACHI'
        }

        support = JciHitachiAWSStatusSupport(raw_he_support)
        assert support.status == processed_he_support
        assert len(processed_he_support) == len(raw_he_support) + 1  # `Brand`

    def test_str2id(self):
        raw_ac_support = {
            'FirmwareId':3,
            'DeviceType':1,
            'Model':'RAD-90NF',
            'FindMe':10,
            'Switch':3,
            'Mode':31,
            'FanSpeed':31,
            'TemperatureSetting':4128,
            'IndoorTemperature':40,
            'SleepModeRemainingTime':1440,
            'MildewProof':3,
            'QuickMode':3,
            'PowerSaving':3,
            'ControlTone':3,
            'PowerConsumption':65535,
            'TaiseiaError':85,
            'FilterElapsedHour':800,
            'CleanSwitch':3,
            'CleanNotification':3,
            'CleanStatus':7,
            'Error':0,
            'FirmwareVersion':'6.0.035',
            'FirmwareCode':35,
            'SystemTimestamp':35000,
            'RequestTimestamp':1648618952,
            'Timestamp':1648618952,
        }
        support = JciHitachiAWSStatusSupport(raw_ac_support)

        # test legacy name and string status
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id("AC", "air_speed", status_str_value="high", support_code=support)
        assert is_valid == True
        assert status_name == "FanSpeed"
        assert status_value == 4

        # test regular name and integer status
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id("AC", "FanSpeed", 4, support_code=support)
        assert is_valid == True
        assert status_name == "FanSpeed"
        assert status_value == 4

        # test incorrect status name
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id("AC", "FanpSpeed", 4, support_code=support)
        assert is_valid == False
    
        # test incorrect status value
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id("AC", "FanSpeed", 6, support_code=support)
        assert is_valid == False


import pytest

from JciHitachi.model import (JciHitachiAC, JciHitachiACSupport, JciHitachiDH,
                              JciHitachiDHSupport, JciHitachiHE,
                              JciHitachiHESupport)
from JciHitachi.status import (JciHitachiCommandAC, JciHitachiCommandDH,
                               JciHitachiCommandHE,
                               JciHitachiStatusInterpreter)

from . import (MOCK_CODE_AC, MOCK_CODE_DH, MOCK_GATEWAY_MAC,
               MOCK_SUPPORT_CODE_AC, MOCK_SUPPORT_CODE_DH, MOCK_COMMAND_AC,
               MOCK_COMMAND_DH)


class TestACStatus:
    def test_model_brand(self):
        dev_support = JciHitachiStatusInterpreter(
            MOCK_SUPPORT_CODE_AC).decode_support()
        ac_support = JciHitachiACSupport(dev_support)

        assert ac_support.brand == "HITACHI"
        assert ac_support.model == "RAD-90NF"
    
    def test_supported_status(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_AC).decode_status()
        ac_status = JciHitachiAC(dev_status).status
        
        for status in ac_status.values():
            assert status != "unknown" and status != None

    def test_unsupported_status(self):
        dev_status = {}
        ac_status = JciHitachiAC(dev_status).status

        for status in ac_status.values():
            assert status == "unsupported" or status == -1

    def test_correctness(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_AC).decode_status()
        ac_status = JciHitachiAC(dev_status).status

        mock_status = {
            'power': 'off',
            'mode': 'cool',
            'air_speed': 'silent',
            'target_temp': 26,
            'indoor_temp': 26,
            'sleep_timer': 0,
            'vertical_wind_swingable': 'unsupported',
            'vertical_wind_direction': -1,
            'horizontal_wind_direction': 'unsupported',
            'mold_prev': 'disabled',
            'fast_op': 'disabled',
            'energy_save': 'disabled',
            'sound_prompt': 'enabled',
            'outdoor_temp': 24,
            'power_kwh': 0.0,
            'freeze_clean': 'off',
        }

        for key, value in ac_status.items():
            assert key in mock_status
            assert mock_status[key] == value

    def test_limit(self):
        dev_support = JciHitachiStatusInterpreter(MOCK_SUPPORT_CODE_AC).decode_support()
        ac_support = JciHitachiACSupport(dev_support)
        mock_status_raw = {
            'power': 0,
            'mode': 0,
            'air_speed': 1,
            'target_temp': 26,
            'indoor_temp': 26,
            'sleep_timer': 0,
            'vertical_wind_swingable': -1,
            'vertical_wind_direction': -1,
            'horizontal_wind_direction': -1,
            'mold_prev': 0,
            'fast_op': 0,
            'energy_save': 0,
            'sound_prompt': 0,
            'outdoor_temp': 24,
            'power_kwh': 0
        }

        mock_status_raw_limited = {key: ac_support.limit(key, value) for key, value in mock_status_raw.items()}

        for key, raw_value in mock_status_raw.items():
            if raw_value == -1:
                raw_value = None
            if key == "outdoor_temp": # The supported `outdoor_temp` values seem not correct so the limited values are incorrect too. skipping this test.
                continue
            assert raw_value == mock_status_raw_limited[key]

    def test_command(self):
        commander = JciHitachiCommandAC(MOCK_GATEWAY_MAC)
        b64command = commander.get_command(MOCK_COMMAND_AC, 0)
        
        mock_command = bytearray.fromhex(
                        "d0d100003c6a9dffff03e0d4ffffffff \
                        00000100000000000000002000010000 \
                        908D9859F3FD7f6c02000d278050f0d4 \
                        469dafd3605a6ebbdb130d278052f0d4 \
                        469dafd3605a6ebbdb13060006019E00 \
                        0099")
        assert b64command == mock_command


class TestDHStatus:
    def test_model_brand(self):
        dev_support = JciHitachiStatusInterpreter(
            MOCK_SUPPORT_CODE_DH).decode_support()
        dh_support = JciHitachiDHSupport(dev_support)

        assert dh_support.brand == "HITACHI"
        assert dh_support.model == "RD-360HH"

    def test_supported_status(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_DH).decode_status()
        dh_status = JciHitachiDH(dev_status).status

        for status in dh_status.values():
            assert status != "unknown" and status != None

    def test_unsupported_status(self):
        dev_status = {}
        ac_status = JciHitachiDH(dev_status).status
        for status in ac_status.values():
            assert status == "unsupported" or status == -1

    def test_correctness(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_DH).decode_status()
        ac_status = JciHitachiDH(dev_status).status

        mock_status = {
            'power': 'on',
            'mode': 'custom',
            'target_humidity': 70,
            'indoor_humidity': 70,
            'wind_swingable': 'on',
            'water_full_warning': 'off',
            'clean_filter_notify': 'disabled',
            'air_purify_level': 'unsupported',  # not implemented
            'air_speed': 'silent',
            'side_vent': 'off',
            'sound_control': 'button+waterfull',
            'error_code': 0,
            'mold_prev': 'off',
            'power_kwh': 0.4,
            'air_quality_value': -1,  # not implemented
            'air_quality_level': 'unsupported',  # not implemented
            'pm25_value': 26,
            'display_brightness': 'bright',
            'odor_level': 'middle',
            'air_cleaning_filter': 'enabled'
        }

        for key, value in ac_status.items():
            assert key in mock_status
            assert mock_status[key] == value

    def test_command(self):
        commander = JciHitachiCommandDH(MOCK_GATEWAY_MAC)
        b64command = commander.get_command(MOCK_COMMAND_DH, 0)

        mock_command = bytearray.fromhex(
            "d0d100003c6a9dffff03e0d4ffffffff \
                        00000100000000000000002000010000 \
                        908D9859F3FD7f6c02000d278050f0d4 \
                        469dafd3605a6ebbdb130d278052f0d4 \
                        469dafd3605a6ebbdb13060006049000 \
                        0092")
        assert b64command == mock_command


class TestCommonStatus:
    def test_command_num_bytes(self):
        ac_commander = JciHitachiCommandAC(MOCK_GATEWAY_MAC)
        assert len(ac_commander.get_command("power", 0)) == 82
        dh_commander = JciHitachiCommandDH(MOCK_GATEWAY_MAC)
        assert len(dh_commander.get_command("power", 0)) == 82
        #he_commander = JciHitachiCommandHE(MOCK_GATEWAY_MAC)
        #assert len(he_commander.get_command("power", 0)) == 82

    # TODO: Add device availablity test